import pytest
from typing import Dict, List
import os

from miya_speechless.models.llm_client import LLMClient, LLMConfig, LLMError
from miya_speechless.nlp.sentiment_analysis import classify_sentiment
from miya_speechless.nlp.loader import render_messages


class MockLLMClient(LLMClient):
    def __init__(self, cfg: LLMConfig, mock_completion):
        super().__init__(cfg)
        self._mock_completion = mock_completion

    def _call_completion(self, *args, **kwargs):
        return self._mock_completion(*args, **kwargs)

    def ask(
        self,
        messages: List[dict],
        *,
        model: str = None,
        temperature: float = None,
        **kwargs,
    ) -> str:
        model_name = model or self.cfg.model
        temp = temperature if temperature is not None else self.cfg.temperature
        try:
            resp = self._call_completion(
                model=f"{self.cfg.provider}/{model_name}",
                messages=messages,
                temperature=temp,
                api_key=self.cfg.api_key,
                **kwargs,
            )
            return resp["choices"][0]["message"]["content"]
        except Exception as exc:
            raise LLMError(
                provider=self.cfg.provider,
                model=model_name,
                message="LLM request failed",
                original=exc,
            )


@pytest.fixture(scope="session")
def cfg() -> LLMConfig:
    return LLMConfig(provider="openai", api_key="test-key", model="gpt-4o")


def _fake_completion_success(*_, **__) -> Dict[str, List[Dict[str, Dict[str, str]]]]:
    return {"choices": [{"message": {"content": "Hello, user!"}}]}


def _fake_completion_error(*_, **__):
    raise RuntimeError("provider blew up")


def test_ask_success(cfg: LLMConfig):
    client = MockLLMClient(cfg, mock_completion=_fake_completion_success)
    result = client.ask(messages=[{"role": "user", "content": "Ping"}])
    assert result == "Hello, user!"


def test_ask_failure_raises_llmerror(cfg: LLMConfig):
    client = MockLLMClient(cfg, mock_completion=_fake_completion_error)

    with pytest.raises(LLMError) as exc:
        client.ask(messages=[{"role": "user", "content": "Ping"}])

    err = exc.value
    assert err.provider == "openai"
    assert err.model == "gpt-4o"
    assert "LLM request failed" in str(err)


@pytest.mark.parametrize("template_name", ["tags", "summarize", "sentiment_score"])
def test_render_template_injects_text(template_name):
    convo = "Hello, this is a dummy conversation."
    messages = render_messages(template_name, transcribed_and_diarized=convo)

    full_content = "\n".join(m["content"] for m in messages)
    assert convo in full_content, f"{template_name}.j2 did not inject the variable"


def test_render_template_keeps_json_format():
    messages = render_messages("summarize", transcribed_and_diarized="dummy")

    full_content = "\n".join(msg["content"] for msg in messages)

    for key in (
        '"reason":',
        '"description":',
        '"products":',
        '"satisfaction":',
        '"fup":',
        '"tone_of_voice":',
        '"tags":',
    ):
        assert key in full_content, f"Missing expected key: {key}"


@pytest.mark.skipif(
    not os.getenv("OPENAI_API_KEY"),
    reason="Requires OPENAI_API_KEY to be set in environment",
)
def test_ask_with_openai_live():
    """Test the real OpenAI client with a real prompt (integration)."""
    cfg = LLMConfig(
        provider="openai",
        api_key=os.environ["OPENAI_API_KEY"],
        model="gpt-4o",
        temperature=0.1,
    )

    client = LLMClient(cfg)

    messages = [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "What is the capital of France?"},
    ]

    response = client.ask(messages)
    assert isinstance(response, str)
    assert "paris" in response.lower()


@pytest.mark.skipif("OPENAI_API_KEY" not in os.environ, reason="No API key in env")
@pytest.mark.parametrize(
    "convo,expected",
    [
        (
            "I'm so happy with the support I received! Everything went smoothly.",
            "Positive",
        ),
        (
            "This was the worst experience I've ever had. Nothing worked.",
            "Negative",
        ),
        (
            "It was okay, not great, not terrible. Just fine.",
            "Neutral",
        ),
    ],
)
def test_sentiment_classification_with_live_openai(convo: str, expected: str):
    """Test full prompt + response + classification using real OpenAI API."""
    cfg = LLMConfig(
        provider="openai",
        api_key=os.environ["OPENAI_API_KEY"],
        model="gpt-4o",
        temperature=0.0,
    )
    client = LLMClient(cfg)

    messages = render_messages("sentiment_score", transcribed_and_diarized=convo)

    raw_score = client.ask(messages)

    try:
        score = float(raw_score.strip())
    except ValueError:
        pytest.fail(f"Expected numeric sentiment score, got: {raw_score}")

    classification = classify_sentiment(score)

    assert (
        classification == expected
    ), f"Expected {expected}, got {classification} (score: {score})"



"""Base class for transcription clients."""

import os
from typing import Any, Dict, Union


class BaseTranscriber:
    """Interface definition for transcription clients."""

    def transcribe(
        self,
        audio_file: Union[str, bytes, os.PathLike],
        *,
        language: str = "en",
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Convert *audio_file* to text.

        Parameters
        ----------
        audio_file : Union[str, bytes, os.PathLike]
            The audio file to transcribe.
        language : str
            The language of the audio file.
        **kwargs : dict
            Additional keyword arguments to pass to the transcription client.

        Raises
        ------
        NotImplementedError
            If the method is not implemented in the subclass.
        """
        raise NotImplementedError



"""Export a pre-trained pyannote model to ONNX format."""

import click
import torch
import onnxruntime as ort

from pyannote.audio import Model

DUMMY_LENGTH = 160000


@click.command()
@click.argument("checkpoint", type=click.Path(exists=True, file_okay=True))
@click.argument("output", type=click.Path(exists=False, file_okay=True))
def main(checkpoint: str, output: str):
    """
    Export an onnx model from a pyannote checkpoint.

    Parameters
    ----------
    checkpoint : str
        The path to the pyannote checkpoint file.
    output : str
        The path to the output ONNX model file.
    """
    model = Model.from_pretrained(checkpoint)

    dummy_input = torch.zeros(3, 1, DUMMY_LENGTH)
    torch.onnx.export(
        model,
        dummy_input,
        output,
        do_constant_folding=True,
        input_names=["input"],
        output_names=["output"],
        dynamic_axes={
            "input": {0: "B", 1: "C", 2: "T"},
        },
    )
    opts = ort.SessionOptions()
    opts.log_severity_level = 3
    opts.optimized_model_filepath = output
    # so.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_ALL
    # ORT_DISABLE_ALL, ORT_ENABLE_BASIC, ORT_ENABLE_EXTENDED, ORT_ENABLE_ALL
    ort.InferenceSession(output, sess_options=opts)


if __name__ == "__main__":
    main()



import io
import types
import os
from typing import Dict

import pytest
import wave
from elevenlabs import ElevenLabs, save
from pydub import AudioSegment

from miya_speechless.transcribe.clients import (
    OpenAITranscriber,
    ElevenLabsTranscriber,
    GoogleTranscriber,
)

DUMMY_WAV = io.BytesIO(b"RIFF....WAVEfmt ")


class _DummyResponse:
    """Mimics the return object of both SDKs."""

    def __init__(self, payload: Dict):
        self._payload = payload

    # OpenAI SDK
    def to_dict(self):  # noqa: D401
        return self._payload

    # ElevenLabs SDK
    def model_dump(self):  # noqa: D401
        return self._payload


def test_openai_transcribe_success(monkeypatch):
    """`transcribe()` should return the payload from `to_dict()` on success."""
    dummy_payload = {"text": "hello from openai"}
    dummy_resp = _DummyResponse(dummy_payload)

    # Stub out client.audio.transcriptions.create
    def _dummy_create(**_):
        return dummy_resp

    dummy_transcriptions = types.SimpleNamespace(create=_dummy_create)
    dummy_audio = types.SimpleNamespace(transcriptions=dummy_transcriptions)
    dummy_client = types.SimpleNamespace(audio=dummy_audio)

    oa = OpenAITranscriber()
    monkeypatch.setattr(oa, "client", dummy_client, raising=False)

    result = oa.transcribe(DUMMY_WAV)
    assert result == dummy_payload


def test_openai_transcribe_failure_returns_empty_dict(monkeypatch):
    """On exception, `transcribe()` should log and return an *empty* dict."""

    def _dummy_create(**_):
        raise RuntimeError("boom")

    dummy_transcriptions = types.SimpleNamespace(create=_dummy_create)
    dummy_audio = types.SimpleNamespace(transcriptions=dummy_transcriptions)
    dummy_client = types.SimpleNamespace(audio=dummy_audio)

    oa = OpenAITranscriber()
    monkeypatch.setattr(oa, "client", dummy_client, raising=False)

    assert oa.transcribe(DUMMY_WAV) == {}


def _elevenlabs_dummy_client(monkeypatch, payload: Dict):
    """Return an `ElevenLabsTranscriber` whose `.convert()` is patched."""
    el = ElevenLabsTranscriber(api_key="test-key")
    dummy_resp = _DummyResponse(payload)

    def _dummy_convert(**_):  # noqa: D401
        return dummy_resp

    dummy_s2t = types.SimpleNamespace(convert=_dummy_convert)
    monkeypatch.setattr(el.client, "speech_to_text", dummy_s2t, raising=False)
    return el


def test_11labs_transcribe_success(monkeypatch):
    payload = {"text": "hello from 11labs"}
    el = _elevenlabs_dummy_client(monkeypatch, payload)
    assert el.transcribe(DUMMY_WAV) == payload


def test_11labs_transcribe_failure_returns_empty_dict(monkeypatch):
    el = ElevenLabsTranscriber(api_key="test-key")

    def _dummy_convert(**_):  # noqa: D401
        raise RuntimeError("boom")

    dummy_s2t = types.SimpleNamespace(convert=_dummy_convert)
    monkeypatch.setattr(el.client, "speech_to_text", dummy_s2t, raising=False)

    assert el.transcribe(DUMMY_WAV) == {}


def _make_silent_wav(path, duration_s: float = 1.0, rate: int = 16_000):
    """Write *duration_s* seconds of 16-kHz 16-bit mono silence to *path*."""
    n_frames = int(duration_s * rate)
    with wave.open(str(path), "wb") as wf:
        wf.setnchannels(1)
        wf.setsampwidth(2)  # 16-bit
        wf.setframerate(rate)
        wf.writeframes(b"\x00\x00" * n_frames)


@pytest.mark.skipif(
    "OPENAI_API_KEY" not in os.environ,
    reason="Requires OPENAI_API_KEY for live Whisper test",
)
def test_openai_integration(tmp_path):
    """Smoke-test Whisper with 1 s of silence (should return an empty string)."""
    dummy_path = tmp_path / "silence.wav"
    _make_silent_wav(dummy_path)

    oa = OpenAITranscriber()
    result = oa.transcribe(dummy_path)

    assert isinstance(result, dict)


@pytest.mark.skipif(
    "ELEVENLABS_API_KEY" not in os.environ,
    reason="Requires ELEVENLABS_API_KEY for live 11Labs test",
)
def test_11labs_integration(tmp_path):
    """Smoke-test ElevenLabs with 1 s of silence (should return an empty string)."""
    dummy_path = tmp_path / "silence.wav"
    _make_silent_wav(dummy_path)

    el = ElevenLabsTranscriber()
    result = el.transcribe(dummy_path)

    assert isinstance(result, dict)


@pytest.mark.skipif(
    "OPENAI_API_KEY" not in os.environ or "ELEVENLABS_API_KEY" not in os.environ,
    reason="Requires both OPENAI_API_KEY and ELEVENLABS_API_KEY",
)
def test_roundtrip_real_tts_to_stt(tmp_path):
    text = "Hello world this is a test from Slovakia with love"
    audio_path = tmp_path / "hello.mp3"

    client = ElevenLabs()

    audio_stream = client.text_to_speech.convert(
        text=text,
        voice_id="EXAVITQu4vr4xnSDxMaL",
        model_id="eleven_multilingual_v2",
        output_format="mp3_44100_128",
    )
    save(audio_stream, str(audio_path))

    openai_text = str(OpenAITranscriber().transcribe(audio_path)).lower()
    el_text = str(ElevenLabsTranscriber().transcribe(audio_path)).lower()

    assert "hello" in openai_text
    assert "hello" in el_text

    assert "slovakia" in openai_text
    assert "slovakia" in el_text


@pytest.mark.integration
def test_google_transcriber_roundtrip_with_real_speech(tmp_path):
    if (
        "GOOGLE_APPLICATION_CREDENTIALS" not in os.environ
        or "GOOGLE_CLOUD_PROJECT" not in os.environ
    ):
        pytest.skip(
            "Set GOOGLE_APPLICATION_CREDENTIALS and GOOGLE_CLOUD_PROJECT for Google STT."
        )

    if "ELEVENLABS_API_KEY" not in os.environ:
        pytest.skip("Set ELEVENLABS_API_KEY for ElevenLabs TTS.")

    text = "Hello world, this is a test from the Miya Speechless pipeline."
    pcm_path = tmp_path / "tts_generated.pcm"

    client = ElevenLabs()
    audio_stream = client.text_to_speech.convert(
        text=text,
        voice_id="EXAVITQu4vr4xnSDxMaL",
        model_id="eleven_multilingual_v2",
        output_format="pcm_16000",  # <-- corrected
    )
    save(audio_stream, str(pcm_path))

    # Convert raw PCM to WAV
    pcm_audio = AudioSegment(
        data=pcm_path.read_bytes(),
        sample_width=2,  # 16-bit PCM = 2 bytes
        frame_rate=16000,
        channels=1,
    )
    wav_path = tmp_path / "tts_generated.wav"
    pcm_audio.export(wav_path, format="wav")

    gt = GoogleTranscriber()
    result = gt.transcribe(wav_path)

    assert isinstance(result, dict), f"Expected dict, got {type(result)}: {result}"
    assert "chunks" in result, f"Expected 'chunks' key in result: {result.keys()}"
    assert isinstance(
        result["chunks"], list
    ), f"'chunks' should be list, got {type(result['chunks'])}"

    found_keyword = False
    for chunk in result["chunks"]:
        if "results" in chunk:
            for res in chunk["results"]:
                for alt in res.get("alternatives", []):
                    transcript = alt.get("transcript", "").lower()
                    if "hello" in transcript and "test" in transcript:
                        found_keyword = True
        elif "speechRecognitionResults" in chunk:
            for res in chunk["speechRecognitionResults"]:
                transcript = res.get("transcript", "").lower()
                if "hello" in transcript and "test" in transcript:
                    found_keyword = True

    assert found_keyword, f"Transcription did not contain expected keywords: {result}"



"""Download and extract audio from a YouTube video as an mp3 file."""

import os
import re
import subprocess  # noqa: S404
from pytubefix import YouTube
from pytubefix.cli import on_progress
import traceback


def clean_filename(title):
    """
    Clean the video title to create a valid filename.

    Parameters
    ----------
    title : str
        The original title of the video.

    Returns
    -------
    str
        A cleaned version of the title suitable for a filename.
    """
    # Convert to lowercase
    title = title.lower()

    # Replace spaces with underscores
    title = title.replace(" ", "_")

    # Remove non-alphanumeric characters except underscores
    return re.sub(r"[^a-z0-9_]", "", title)


# There is a bug with pytube that causes an error when downloading audio streams
# Therefore, we use pytubefix, which is a fork of pytube that fixes this issue
def download_and_extract_audio(video_url, output_path="data/audio", overwrite=False):
    """
    Download a YouTube video and extract the audio as a mp3 file.

    Parameters
    ----------
    video_url : str
        The URL of the YouTube video to download.
    output_path : str
        The directory where the audio file will be saved (default is 'data/audio').
    overwrite : bool
        Whether to overwrite the audio file if it already exists (default is False).

    Raises
    ------
    RuntimeError
        If the ffmpeg conversion fails.
    ValueError
        If no audio stream is available for the video.
    """
    try:
        # Create output directory if it doesn't exist
        os.makedirs(output_path, exist_ok=True)

        # Create a YouTube object
        yt = YouTube(video_url, on_progress_callback=on_progress)
        cleaned_title = clean_filename(yt.title)

        # Get the audio stream (check available streams instead of using first)
        audio_stream = yt.streams.filter(only_audio=True, file_extension="mp4").first()
        if not audio_stream:
            raise ValueError("No audio stream available.")

        temp_audio_file_path = os.path.join(output_path, f"{cleaned_title}.mp4")
        final_mp3_file_path = os.path.join(output_path, f"{cleaned_title}.mp3")

        # Check if the file already exists and handle overwriting
        if os.path.exists(final_mp3_file_path):
            if overwrite:
                print(f"File {final_mp3_file_path} exists. Overwriting.")
                os.remove(final_mp3_file_path)
            else:
                print(f"File {final_mp3_file_path} already exists. Skipping download.")
                return

        # Download the audio file as .mp4
        print(f"Downloading audio from: {yt.title}")
        audio_stream.download(output_path=output_path, filename=f"{cleaned_title}.mp4")

        # Convert the downloaded file to mp3 using ffmpeg with the -y flag to automatically overwrite
        command = [
            "ffmpeg",
            "-y",
            "-i",
            temp_audio_file_path,
            "-vn",
            "-acodec",
            "libmp3lame",
            final_mp3_file_path,
        ]

        result = subprocess.run(  # noqa: S603
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True
        )

        if result.returncode != 0:
            error_message = result.stderr.decode()
            raise RuntimeError(f"ffmpeg conversion failed: {error_message}")

        # Remove the temporary .mp4 file after conversion
        os.remove(temp_audio_file_path)

        print(f"Audio extracted and saved as: {final_mp3_file_path}")

    except Exception as e:
        print(f"An error occurred: {e}")
        print(traceback.format_exc())



"""Transcription clients for OpenAI and ElevenLabs."""

import os
import logging
import math
import base64
import time
from io import BytesIO
from pathlib import Path
from typing import Any, Dict, Union
from pydub import AudioSegment

from openai import OpenAI
from elevenlabs.client import ElevenLabs
from google.cloud.speech_v2 import SpeechClient
from google.cloud.speech_v2.types import cloud_speech
from google.oauth2 import service_account
from google.protobuf import json_format

from miya_speechless.transcribe.base import BaseTranscriber

logger = logging.getLogger(__name__)


class OpenAITranscriber(BaseTranscriber):
    """Client wrapper around the **OpenAI Whisper** transcription API."""

    def __init__(self, model: str = "whisper-1") -> None:
        """
        Initialize the OpenAI transcription client.

        Parameters
        ----------
        model : str
            The model to use for transcription.
        """
        self.client = OpenAI()
        self.model = model

    def transcribe(
        self,
        audio_file: Union[str, bytes, os.PathLike],
        *,
        language: str = "en",
        temperature: float = 0,
    ) -> Dict[str, Any]:
        """
        Transcribe an audio file using the OpenAI API.

        Parameters
        ----------
        audio_file : Union[str, bytes, os.PathLike]
            The audio file to transcribe.
        language : str
            The language of the audio file.
        temperature : float
            The temperature of the transcription.

        Returns
        -------
        dict
            The transcription of the audio file.
        """
        response_format = "json" if "gpt" in self.model else "verbose_json"

        try:
            transcript = self.client.audio.transcriptions.create(
                file=audio_file,
                model=self.model,
                response_format=response_format,
                timestamp_granularities=["segment"],
                language=language,
                temperature=temperature,
            )
            return transcript.to_dict()
        except Exception as exc:
            logger.error("OpenAI transcription failed: %s", exc)
            return {}


class ElevenLabsTranscriber(BaseTranscriber):
    """Client wrapper around **ElevenLabs Scribe** speech‑to‑text."""

    def __init__(
        self,
        *,
        model_id: str = "scribe_v1",
        tag_audio_events: bool = True,
        diarize: bool = False,
        api_key: str | None = None,
    ) -> None:
        """
        Initialize the ElevenLabs transcription client.

        Parameters
        ----------
        model_id : str
            The model to use for transcription.
        tag_audio_events : bool
            Whether to tag audio events.
        diarize : bool
            Whether to diarize the audio.
        api_key : str, optional
            The API key to use for the ElevenLabs transcription client.
        """
        self.client = ElevenLabs(api_key=api_key or os.getenv("ELEVENLABS_API_KEY"))
        self.model_id = model_id
        self.tag_audio_events = tag_audio_events
        self.diarize = diarize

    def transcribe(
        self,
        audio_file: Union[str, bytes, os.PathLike],
        *,
        language: str = "eng",
    ) -> Dict[str, Any]:
        """
        Transcribe an audio file using the ElevenLabs API.

        Parameters
        ----------
        audio_file : Union[str, bytes, os.PathLike]
            The audio file to transcribe.
        language : str
            The language of the audio file.
        """
        if isinstance(audio_file, (str, Path, os.PathLike)):
            with open(audio_file, "rb") as f:
                audio_bytes = f.read()
        elif hasattr(audio_file, "read"):
            audio_bytes = audio_file.read()
            audio_file.seek(0)
        else:
            audio_bytes = audio_file  # assume ``bytes``

        audio_io = BytesIO(audio_bytes)

        try:
            result = self.client.speech_to_text.convert(
                file=audio_io,
                model_id=self.model_id,
                language_code=language,
                tag_audio_events=self.tag_audio_events,
                diarize=self.diarize,
            )
        except Exception as exc:
            logger.error("ElevenLabs transcription failed: %s", exc)
            return {}

        if hasattr(result, "model_dump"):
            return result.model_dump()
        if hasattr(result, "to_dict"):
            return result.to_dict()
        return result


class GoogleTranscriber(BaseTranscriber):
    """Client wrapper around **Google Cloud Speech-to-Text v2**."""

    def __init__(
        self,
        *,
        project_id: str | None = None,
        cred_path: str | None = None,
        location: str = "europe-west4",
        language_code: str = "en-US",
        chunk_length_ms: int = 59990,
    ) -> None:
        """
        Initialize the Google Cloud transcription client.

        Parameters
        ----------
        project_id : str, optional
            GCP Project ID.
        cred_path : str, optional
            Path to GCP service account JSON.
        location : str
            GCP region for Speech-to-Text.
        language_code : str
            Language code for transcription.
        chunk_length_ms : int
            Length of each chunk in milliseconds for local transcription.
        """
        cred_path = cred_path or os.getenv("GOOGLE_APPLICATION_CREDENTIALS")
        self.project_id = project_id or os.getenv("GOOGLE_CLOUD_PROJECT")
        self.credentials = service_account.Credentials.from_service_account_file(
            cred_path
        )
        self.client = SpeechClient(
            credentials=self.credentials,
            client_options={"api_endpoint": f"{location}-speech.googleapis.com"},
        )
        self.location = location
        self.language_code = language_code
        self.chunk_length_ms = chunk_length_ms

    def transcribe(
        self,
        audio_file: Union[str, bytes, os.PathLike],
        *,
        language: str = "en-US",
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Transcribe an audio file using Google Cloud Speech-to-Text.

        Parameters
        ----------
        audio_file : Union[str, bytes, os.PathLike]
            Local WAV path or GCS URI.
        language : str
            Language code for transcription.

        Returns
        -------
        dict
            Transcription results.
        """
        self.language_code = language

        try:
            if isinstance(audio_file, str) and audio_file.startswith("gs://"):
                logger.info(f"Starting GCS batch transcription for {audio_file}")
                return self._transcribe_batch(audio_file)
            elif isinstance(audio_file, (str, Path, os.PathLike)):
                logger.info(f"Starting local chunked transcription for {audio_file}")
                return self._transcribe_local(str(audio_file))

            logger.error(
                "Unsupported input for GoogleTranscriber; requires file path or GCS URI."
            )
            return {}
        except Exception as exc:
            logger.error("Google Cloud transcription failed: %s", exc)
            return {}

    def _make_config(self) -> cloud_speech.RecognitionConfig:
        """Make a recognition config for the Google Cloud Speech-to-Text API."""
        return cloud_speech.RecognitionConfig(
            auto_decoding_config=cloud_speech.AutoDetectDecodingConfig(),
            language_codes=[self.language_code],
            model="chirp_2",
            features=cloud_speech.RecognitionFeatures(
                enable_automatic_punctuation=True,
                profanity_filter=False,
                enable_word_time_offsets=True,
                multi_channel_mode="SEPARATE_RECOGNITION_PER_CHANNEL",
                max_alternatives=1,
            ),
        )

    def _transcribe_chunk(self, chunk: AudioSegment) -> dict:
        """
        Transcribe a chunk of audio using the Google Cloud Speech-to-Text API.

        Parameters
        ----------
        chunk : AudioSegment
            The chunk of audio to transcribe.

        Returns
        -------
        dict
            The transcription of the chunk.
        """
        buffer = BytesIO()
        chunk.export(buffer, format="wav", parameters=["-acodec", "pcm_s16le"])
        buffer.seek(0)
        audio_content = base64.b64encode(buffer.read()).decode("utf-8")

        request = cloud_speech.RecognizeRequest(
            recognizer=f"projects/{self.project_id}/locations/{self.location}/recognizers/_",
            config=self._make_config(),
            content=audio_content,
        )
        response = self.client.recognize(request=request)
        return json_format.MessageToDict(response._pb)

    def _transcribe_local(self, audio_path: str) -> dict:
        """
        Transcribe a local audio file using the Google Cloud Speech-to-Text API.

        Parameters
        ----------
        audio_path : str
            The path to the local audio file to transcribe.

        Returns
        -------
        dict
            The transcription of the local audio file.
        """
        audio = AudioSegment.from_wav(audio_path)
        total_chunks = math.ceil(len(audio) / self.chunk_length_ms)
        results = []

        for i in range(total_chunks):
            start = i * self.chunk_length_ms
            end = min((i + 1) * self.chunk_length_ms, len(audio))
            chunk = audio[start:end]
            logger.debug(
                f"Transcribing chunk {i+1}/{total_chunks} ({start}ms - {end}ms)"  # noqa: WPS237
            )
            result = self._transcribe_chunk(chunk)
            results.append(result)

        return {"chunks": results}

    def _transcribe_batch(self, gcs_uri: str) -> dict:
        """
        Transcribe a batch of audio files using the Google Cloud Speech-to-Text API.

        Parameters
        ----------
        gcs_uri : str
            The GCS URI of the audio file to transcribe.

        Returns
        -------
        dict
            The transcription of the batch of audio files.
        """
        config = self._make_config()
        audio_metadata = cloud_speech.BatchRecognizeFileMetadata(uri=gcs_uri)
        batch_request = cloud_speech.BatchRecognizeRequest(
            config=config,
            files=[audio_metadata],
            recognition_output_config=cloud_speech.RecognitionOutputConfig(
                inline_response_config=cloud_speech.InlineOutputConfig()
            ),
            recognizer=f"projects/{self.project_id}/locations/{self.location}/recognizers/_",
        )
        operation = self.client.batch_recognize(request=batch_request)
        while not operation.done():
            logger.debug("Waiting for batch transcription to complete...")
            time.sleep(2)
        response = operation.result()
        return json_format.MessageToDict(response._pb)



"""OpenAI API client for the Scout app."""

from dataclasses import dataclass
from litellm import completion
from typing import List, Dict, Any
from tenacity import retry, wait_exponential, stop_after_attempt

TIMEOUT_SECONDS = 30


class LLMError(Exception):
    """Raised when an LLM request ultimately fails."""

    def __init__(
        self,
        provider: str,
        model: str,
        message: str,
        *,
        status: int = None,
        original: Exception = None,
    ):
        """
        Initialize the LLMError with the specified provider, model, and message.

        Args:
            provider (str): The provider of the LLM.
            model (str): The model of the LLM.
            message (str): The message to display.
            status (int | None): The status code of the error.
            original (Exception | None): The original exception.
        """
        super().__init__(message)
        self.provider = provider
        self.model = model
        self.status = status
        self.original = original

    def __str__(self) -> str:
        """This is used to display the error in the logs"""
        core = f"[{self.provider}:{self.model}] {self.args[0]}"
        return f"{core} (status={self.status})" if self.status else core


@dataclass(frozen=True)
class LLMConfig:
    """Configuration for the LLMClient."""

    provider: str
    api_key: str
    model: str = "gpt-4o"
    temperature: float = 0.1


class LLMClient:
    """Tiny wrapper around LiteLLM `completion()` with retries & defaults."""

    def __init__(self, cfg: LLMConfig) -> None:
        """
        Initialize the LLMClient with the specified configuration.

        Args:
            cfg (LLMConfig): The configuration for the LLMClient.
        """
        self.cfg = cfg

    @retry(
        wait=wait_exponential(multiplier=1, max=10),
        stop=stop_after_attempt(4),
        reraise=True,
    )
    def ask(
        self,
        messages: List[Dict[str, str]],
        *,
        model: str = None,  # noqa: DAR401
        temperature: float = None,
        **kwargs: Any,
    ) -> str:
        """
        Send a chat completion request and return the assistant’s text.

        Args:
            messages (List[Dict[str, str]]): The messages to send to the LLM.
            model (str | None): The model to use for completion.
            temperature (float | None): The sampling temperature.
            **kwargs: Additional keyword arguments to pass to the LiteLLM completion function.

        Returns:
            str: The response from the LLM.

        Raises:
            LLMError: If the LLM request fails.
        """
        model_name = model or self.cfg.model
        temp = temperature if temperature is not None else self.cfg.temperature

        provider_name = self.cfg.provider
        full_model_name = f"{provider_name}/{model_name}"
        try:
            resp = completion(
                model=full_model_name,
                messages=messages,
                temperature=temp,
                api_key=self.cfg.api_key,
                **kwargs,
            )
            return resp["choices"][0]["message"]["content"]
        except Exception as exc:
            raise LLMError(
                provider=self.cfg.provider,
                model=model_name,
                message="LLM request failed",
                original=exc,
            )



"""Prompt class for generating prompts"""

import json


class Prompts:
    """Class for generating prompts"""

    @staticmethod
    def get_tags_prompt(transcribed_and_diarized: str) -> tuple:
        """
        Generate a prompt for getting tags from OpenAI.

        Parameters
        ----------
        transcribed_and_diarized : str
            The transcribed and diarized conversation.

        Returns
        -------
        tuple
            A tuple containing the system message and user message.
        """
        system_message = "You are a helpful analysis of transactional sales data."
        user_message = f"Return a comma-separated list of maximum five tags for the following conversation. \n\n Conversation: {transcribed_and_diarized}"
        return system_message, user_message

    @staticmethod
    def summarize_conversation_prompt(transcribed_and_diarized: str) -> tuple:
        """
        Generate a prompt for summarizing a conversation.

        Parameters
        ----------
        transcribed_and_diarized : str
            The transcribed and diarized conversation.

        Returns
        -------
        tuple
            A tuple containing the system message and user message.
        """
        system_message = "You are a helpful assistant for summarizing conversations."
        # user_message = f"Summarize the following conversation in a concise manner. Clearly state the objectives for all speakers. \n\n Conversation: {transcribed_and_diarized}"

        user_message = f"""
        Conversation:
        {transcribed_and_diarized}
        Your task is to create a summary of the presented conversation. If possible, the summary should include:
            - The reason for the customer's contact with support.
            - A brief description of the resolution process and the outcome.
            - The company's products and any offers made by the operator to the customer.
            - A one-sentence evaluation of the customer's satisfaction.
            - If follow-up call necessary, then True, else False.
            - Tone of voice on the scale from 1 to 5 for these pairs, assigning 1 more close to the first pair and vice versa
                1. formal -> casual
                2. serious -> funny
                3. respectfully -> irreverent
                4. matter of fact -> enthusiastic
            - A comma-separated list of maximum five tags for the following conversation.

         Conversely, the summary should not include:
            - Personal information of the customer and operator (e.g., names, addresses).
            - Greetings exchanged between the customer and operator.

        Return result as simple JSON:
            {{
                "reason": "",
                "description": "",
                "products": [""],
                "satisfaction": "",
                "fup": "",
                "tone_of_voice": [
                    {{"formal_casual": ""}},
                    {{"serious_funny": ""}},
                    {{"respect_irreverent": ""}},
                    {{"mof_enthusiastic": ""}}
                ],
                "tags": [""]
            }}
        """  # noqa: E101, W191
        return system_message, user_message

    @staticmethod
    def get_sentiment_score_prompt(transcribed_and_diarized: str) -> tuple:
        """
        Generate a prompt for getting the sentiment score of a conversation.

        Parameters
        ----------
        transcribed_and_diarized : str
            The transcribed and diarized conversation.

        Returns
        -------
        tuple
            A tuple containing the system message and user message.
        """
        system_message = (
            "Provide a floating-point representation of the sentiment of "
            + "the following customer product review that is "
            + "rounded to five decimal places. "
            + "The scale ranges from -1.0 (negative) to 1.0 (positive) "
            + "where 0.0 represents neutral sentiment. "
            + "Only return the sentiment score value and nothing else. "
            + "You may only return a single numeric value."
        )
        user_message = f"Conversation: {transcribed_and_diarized}"
        return system_message, user_message

    @staticmethod
    def parse_summary_response(raw_response: str) -> dict:
        """
        Parse OpenAI's response, clean code block markers, and convert to JSON.

        Parameters
        ----------
        raw_response : str
            The raw response from OpenAI.

        Returns
        -------
        dict
            Parsed JSON dictionary.

        Raises
        ------
        ValueError
            If JSON decoding fails or required keys are missing.
        """
        lines = raw_response.splitlines()
        json_lines = [line for line in lines if not line.startswith("```")]
        json_str = "\n".join(json_lines)

        try:
            summary_dict = json.loads(json_str)
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON format: {e}")

        # Validate required keys
        required_keys = [
            "reason",
            "description",
            "products",
            "satisfaction",
            "fup",
            "tone_of_voice",
            "tags",
        ]
        for key in required_keys:
            if key not in summary_dict:
                raise ValueError(f"Missing expected key in summary: '{key}'")

        return summary_dict



"""
This module provides a utility function to convert audio files from one format to another using the `pydub` library.

The function `convert_audio` allows for easy conversion between
different audio formats by specifying an input file, an output file,
and an optional output format.
"""

from pydub import AudioSegment
import os
from typing import Optional


def convert_audio(
    input_path: str, output_path: str, output_format: Optional[str] = None
) -> None:
    """
    Convert an audio file from any supported format to another format.

    Parameters
    ----------
    input_path : str
        Path to the input audio file.
    output_path : str
        Path where the output audio file will be saved.
        The file extension determines the output format unless `output_format` is specified.
    output_format : Optional[str], default=None
        Desired output format (e.g., 'wav', 'flac', 'mp3').
        If not provided, the format is inferred from the `output_path` extension.

    Raises
    ------
    FileNotFoundError
        If the input file does not exist.
    ValueError
        If the input or output file does not have a valid extension to determine its format.

    Examples
    --------
    Convert MP3 to WAV:

    >>> convert_audio("example.mp3", "example_converted.wav")

    Convert M4A to FLAC:

    >>> convert_audio("example.m4a", "example_converted.flac")

    Convert WAV to MP3 by specifying the output format:

    >>> convert_audio("example.wav", "example_converted.mp3", output_format="mp3")

    Convert any supported format to another by ensuring correct extensions:

    >>> convert_audio("example.ogg", "example_converted.aac")
    """
    try:
        # Validate input file existence
        if not os.path.isfile(input_path):
            raise FileNotFoundError(f"Input file '{input_path}' does not exist.")

        # Infer input format from the file extension
        input_extension = os.path.splitext(input_path)[1]
        if not input_extension:
            raise ValueError("Input file must have a valid extension.")
        input_format = input_extension[1:].lower()

        # Determine output format
        if output_format:
            fmt = output_format.lower()
            # Ensure output_path has the correct extension
            output_extension = os.path.splitext(output_path)[1]
            if output_extension.lower() != f".{fmt}":
                output_path = os.path.splitext(output_path)[0] + f".{fmt}"
        else:
            output_extension = os.path.splitext(output_path)[1]
            if not output_extension:
                raise ValueError("Output file must have a valid extension.")
            fmt = output_extension[1:].lower()

        # Load the audio file with the inferred input format
        audio = AudioSegment.from_file(input_path, format=input_format)

        # Export the audio in the desired format
        audio.export(output_path, format=fmt)
        print(f"Successfully converted '{input_path}' to '{output_path}'")

    except (FileNotFoundError, ValueError) as e:
        print(f"Error converting '{input_path}': {e}")



"""Jinja2 template loader."""

from importlib.resources import files
from jinja2 import Environment, BaseLoader, select_autoescape
from typing import List, Dict

# Jinja2 environment
env = Environment(
    loader=BaseLoader(),
    autoescape=select_autoescape(default_for_string=True),
    trim_blocks=True,
    lstrip_blocks=True,
)


def render_template(template_name: str, **context) -> str:
    """
    Render a Jinja2 template from `miya_speechless/prompts/`.

    Parameters
    ----------
    template_name (str): Filename without `.j2` extension (e.g., 'summarize').
    **context: Variables to substitute into the template.

    Returns
    -------
    str: Rendered text from the template.
    """
    template_path = files("miya_speechless.prompts").joinpath(f"{template_name}.j2")
    template_str = template_path.read_text(encoding="utf-8")
    template = env.from_string(template_str)
    return template.render(**context)


def render_messages(template_name: str, **context) -> List[Dict[str, str]]:
    """
    Render a template and return OpenAI-style message list with roles.

    Parameters
    ----------
    template_name (str): Filename without `.j2` extension (e.g., 'summarize').
    **context: Variables to substitute into the template.

    Returns
    -------
    List[Dict[str, str]]: Messages suitable for OpenAI Chat API.

    Raises
    ------
    ValueError: If the template does not contain `System:` and `User:` blocks.
    """  # noqa: DAR401
    full_text = render_template(template_name, **context)
    try:
        system_text, user_text = full_text.split("User:")
    except ValueError as e:
        raise ValueError(f"Template '{template_name}.j2' must contain 'User:'") from e

    return [
        {"role": "system", "content": system_text.replace("System:", "").strip()},
        {"role": "user", "content": user_text.strip()},
    ]



"""ONNX utilities for speechless."""

import onnxruntime as ort
from itertools import permutations
import librosa
import numpy as np
from tqdm import tqdm
from typing import Union


INITIAL_SAMPLE_OFFSET = 0
FRAME_STEP_SIZE = 270
NUM_SPEAKERS = 2


class PickableInferenceSession:
    """
    A picklable wrapper around the ONNX InferenceSession.
    """

    def __init__(self, model_path: str):
        """
        Initializes the ONNX InferenceSession.

        Parameters
        ----------
        model_path : str
            The path to the ONNX model file.
        """
        self.model_path = model_path
        self.sess = self._init_session(self.model_path)

    def __getstate__(self):
        """Prepares the object for pickling by only storing the model path."""
        return {"model_path": self.model_path}

    def __setstate__(self, state):
        """Restores the object from a pickled state."""
        self.model_path = state["model_path"]
        self.sess = self._init_session(self.model_path)

    def run(self, *args):
        """Forward method for running inference."""
        return self.sess.run(*args)

    def _init_session(self, model_path: str) -> ort.InferenceSession:
        """
        Initializes the ONNX Runtime session.

        Parameters
        ----------
        model_path : str
            The path to the ONNX model file.
        """
        opts = ort.SessionOptions()
        opts.inter_op_num_threads = 1
        opts.intra_op_num_threads = 1
        opts.log_severity_level = 3
        return ort.InferenceSession(model_path, sess_options=opts)


class PyannoteONNX:
    """Pyannote Speaker Diarization using ONNX model"""

    def __init__(
        self,
        sample_rate: int,
        path_to_onnx_model: str = "segmentation-3.0.onnx",
        show_progress: bool = False,
    ):
        """
        Initialize Pyannote Speaker Diarization using ONNX model

        Parameters
        ----------
        sample_rate : int
            Sample rate of th
        path_to_onnx_model : str
            Path to ONNX model
        show_progress : bool
            Show progress bar, by default False
        """
        self.sample_rate = sample_rate
        self.show_progress = show_progress
        self.num_speakers = NUM_SPEAKERS

        self.session = PickableInferenceSession(path_to_onnx_model)

    def __call__(self, x, step=None, return_chunk=False):
        """
        Perform speaker diarization

        Parameters
        ----------
        x : np.ndarray
            Audio waveform
        step : int, optional
            Step size, by default None
        return_chunk : bool, optional
            Return chunk, by default False

        Yields
        ------
        np.ndarray
            Speaker probabilities
        """
        duration_multiplier = 0.9
        # Estimate the duration from the input
        duration = self.estimate_duration(x)

        if step is None:
            step = duration // 2
        else:
            step = int(step * self.sample_rate)
        step = max(min(step, duration_multiplier * duration), duration // 2)

        windows = list(self.sliding_window(x, duration, step))
        if self.show_progress:
            progress_bar = tqdm(
                total=len(windows),
                desc="Pyannote processing",
                unit="frames",
                bar_format="{l_bar}{bar}{r_bar} | {percentage:.2f}%",
            )

        overlap = self.sample2frame(duration - step)
        overlap_chunk = np.zeros((overlap, self.num_speakers))

        for idx, (window_size, window) in enumerate(windows):
            if self.show_progress:
                progress_bar.update(1)

            ort_outs = self.session.run(None, {"input": window[None, None, :]})[0][0]

            # Exponentiate the outputs to interpret them as probabilities
            ort_outs = np.exp(ort_outs)

            # Aggregate probabilities for speaker #1 and speaker #2 only
            ort_outs[:, 1] += ort_outs[
                :, 4
            ]  # Add overlap between speaker #1 and speaker #2 to speaker #1

            ort_outs[:, 2] += ort_outs[
                :, 4
            ]  # Add overlap between speaker #1 and speaker #2 to speaker #2

            # Select only the relevant columns (speaker #1 and speaker #2)
            ort_outs = ort_outs[:, [1, 2]]

            ort_outs = self.reorder(overlap_chunk, ort_outs)
            if idx != 0:
                ort_outs[:overlap, :] = (ort_outs[:overlap, :] + overlap_chunk) / 2
            if idx != len(windows) - 1:  # noqa: WPS504
                overlap_chunk = ort_outs[-overlap:, :]
                ort_outs = ort_outs[:-overlap, :]
            else:
                ort_outs = ort_outs[: self.sample2frame(window_size), :]

            if return_chunk:
                yield ort_outs
            else:
                for out in ort_outs:  # noqa: WPS526
                    yield out

    @staticmethod
    def sample2frame(x):
        """
        Sample to frame conversion

        Parameters
        ----------
        x : int
            Sample value

        Returns
        -------
        int
            Frame value
        """
        return (x - INITIAL_SAMPLE_OFFSET) // FRAME_STEP_SIZE

    @staticmethod
    def frame2sample(x):
        """
        Frame to sample conversion

        Parameters
        ----------
        x : int
            Frame value

        Returns
        -------
        int
            Sample value
        """
        return (x * FRAME_STEP_SIZE) + INITIAL_SAMPLE_OFFSET

    @staticmethod
    def sliding_window(waveform, window_size, step_size):
        """
        Slide window over waveform

        Parameters
        ----------
        waveform : np.ndarray
            Waveform to slide window over
        window_size : int
            Window size
        step_size : int
            Step size

        Yields
        ------
        Tuple[int, np.ndarray]
            Window size and window data
        """
        windows = []
        start = 0
        num_samples = len(waveform)
        while start <= num_samples - window_size:
            windows.append((start, start + window_size))
            yield window_size, waveform[start : start + window_size]
            start += step_size
        # last incomplete window
        if num_samples < window_size or (num_samples - window_size) % step_size > 0:
            last_window = waveform[start:]
            last_window_size = len(last_window)
            if last_window_size < window_size:
                last_window = np.pad(last_window, (0, window_size - last_window_size))
            yield last_window_size, last_window

    @staticmethod
    def reorder(x, y):
        """
        Reorder the speakers and aggregate

        Parameters
        ----------
        x : np.ndarray
            Overlap chunk
        y : np.ndarray
            Speaker probabilities

        Returns
        -------
        np.ndarray
            Reordered speaker probabilities
        """
        perms = [np.array(perm).T for perm in permutations(y.T)]
        diffs = np.sum(
            np.abs(
                np.sum(np.array(perms)[:, : x.shape[0], :] - x, axis=1)
            ),  # noqa: WPS221
            axis=1,
        )
        return perms[np.argmin(diffs)]

    def estimate_duration(self, x):
        """
        Estimate the duration dynamically based on the input waveform

        Parameters
        ----------
        x : np.ndarray
            Audio waveform

        Returns
        -------
        int
            Estimated duration in samples
        """
        duration_in_seconds = len(x) / self.sample_rate
        estimated_duration = min(
            max(5, duration_in_seconds // 2), 7
        )  # Clamp between 5 and 7 seconds
        return int(estimated_duration * self.sample_rate)

    def itertracks(
        self, wav: Union[str, np.ndarray], onset: float = 0.5, offset: float = 0.5
    ):
        """
        Perform speaker diarization

        Parameters
        ----------
        wav : str or np.ndarray
            Audio waveform
        onset : float
            Onset threshold, by default 0.5
        offset : float
            Offset threshold, by default 0.5

        Yields
        ------
        Dict[str, Union[int, float]]
            Speaker tag, start second, end second
        """
        if not isinstance(wav, np.ndarray):
            wav, _ = librosa.load(wav, sr=self.sample_rate, mono=True)

        current_samples = INITIAL_SAMPLE_OFFSET  # Start with the initial offset

        is_active = [False] * (self.num_speakers)
        start = [0] * (self.num_speakers)

        # Iterate over the speaker probabilities
        for speech_probs in self(wav):
            current_samples += FRAME_STEP_SIZE  # Use FRAME_STEP_SIZE for incrementing
            for idx, prob in enumerate(speech_probs[: self.num_speakers]):

                if is_active[idx]:
                    if prob < offset:
                        yield {  # noqa: WPS220
                            "speaker_tag": idx,
                            "start_sec": round(start[idx] / self.sample_rate, 3),
                            "end_sec": round(current_samples / self.sample_rate, 3),
                        }
                        is_active[idx] = False  # noqa: WPS220
                else:
                    if prob > onset:  # noqa: WPS513
                        start[idx] = current_samples  # noqa: WPS220
                        is_active[idx] = True  # noqa: WPS220

        for idx in range(self.num_speakers):  # noqa: WPS440
            if is_active[idx]:
                yield {  # noqa: WPS220
                    "speaker_tag": idx,
                    "start_sec": round(start[idx] / self.sample_rate, 3),
                    "end_sec": round(current_samples / self.sample_rate, 3),
                }


def diarize_audio(
    audio_path: str,
    sample_rate: int = 16000,
    onset_threshold: float = 0.01,
    offset_threshold: float = 0.01,
    path_to_onnx_model: str = "segmentation-3.0.onnx",
):
    """
    Perform speaker diarization using PyannoteONNX and optionally plot VAD probabilities.

    Parameters
    ----------
    audio_path : str
        Path to the input audio file.
    sample_rate : int
        The sample rate to load the audio with, matching the model's expected rate.
    path_to_onnx_model : str
        Path to the ONNX model file.

    Returns
    -------
    None
    """
    # Initialize the PyannoteONNX model
    pyannote = PyannoteONNX(
        sample_rate,
        path_to_onnx_model,
    )
    output = []

    # Load the audio file as a waveform
    wav, sr = librosa.load(audio_path, sr=sample_rate)

    # Perform diarization and print each detected segment
    for turn in pyannote.itertracks(
        wav, onset=onset_threshold, offset=offset_threshold
    ):
        output.append(turn)

    return output



"""Configuration of default global parameters of the app"""

SAMPLING_RATE = 44100
TEMP_RESULTS_DIR = "data/temp_results"
PATH_TO_ONNX_MODEL = "models/segmentation-3.0.onnx"

# Open AI API transcription models
OPENAI_TRANSCRIPTION_MODELS = [
    "whisper-1",
    "gpt-4o-mini-transcribe",
    "gpt-4o-transcribe",
]

# Real-time transcription defaults
RTT_URL = "wss://api.openai.com/v1/realtime?intent=transcription"
PING_INTERVAL = 5
PING_TIMEOUT = 20
SEND_INTERVAL = 0.01
FRAMES_PER_BUFFER = 3200
CHANNELS = 1
RATE = 16000
